import dataclasses
import typing


@dataclasses.dataclass(frozen=True)
class CameraInfo:
    loaded: bool = False

    model: str = None  # TODO: Use an enum instead
    firmware_version: str = None

    def copy(self, **replacements):
        replacements.setdefault("loaded", True)
        return dataclasses.replace(self, **replacements)


@dataclasses.dataclass(frozen=True)
class CameraFileList:
    uri: typing.List[str]
    totalCount: int
    response_code: int
    message_code: int



from typing import Optional, Coroutine

import asyncio
import logging
import threading
from collections import defaultdict
from uuid import uuid4


class EventManager:
    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
    ):
        self.logger = logger or logging.getLogger(__name__)

        self._registry = defaultdict(dict)
        self._event_loop = asyncio.new_event_loop()
        self._loop_thread = threading.Thread(
            target=self._run_event_loop,
            daemon=True,
        )
        self._started = False

    def _run_event_loop(self):
        # Set the loop for the current thread
        asyncio.set_event_loop(self._event_loop)
        self._event_loop.run_forever()

    def _ensure_started(self):
        if self._started:
            return

        self._loop_thread.start()
        self._started = True

    def register_handler(
        self,
        event_name: str,
        handler: Coroutine,
        uid: Optional[str] = None,
        wait: bool = False,
    ):
        """
        Registers a coroutine handler to be called when an event with
        `event_name` is emitted. `uid` may be explicitly provided, otherwise
        will be randomly generated.

        Parameters:
            event_name: Event that the handler will be called upon
            handler: Handler that will be called upon an event
            uid: A unique identifier for a signal receiver in cases where
                 duplicate handlers may be registered.
            wait: If `True`, the event loop will wait until the handler is
                  executed when invoked.

        Returns:
            None
        """
        if not uid:
            uid = "{}-{}".format(handler.__name__, uuid4())

        # Delay thread start until the first event is registered, as there is no
        # point running a thread until it's actually needed
        self._ensure_started()

        self._registry[event_name][uid] = (handler, wait)
        return uid

    def unregister_handler(self, event_name: str, uid: str):
        handler = self._registry[event_name].pop(uid, None)
        removed_handler = handler is not None
        return removed_handler

    def process_event(self, event_name: str, **kwargs):
        handlers = self._registry.get(event_name, {})
        if not handlers:
            return False

        for uid, (event_handler, should_wait) in handlers.items():
            try:
                coro = event_handler(**kwargs)
            except TypeError:
                self.logger.exception(
                    'Error creating a handler coroutine "{}" for event "{}"'.format(
                        event_handler.__name__,
                        event_name,
                    ),
                )
                continue

            future = None
            try:
                future = asyncio.run_coroutine_threadsafe(
                    coro,
                    self._event_loop,
                )
            except Exception as e:
                self.logger.exception(
                    'Error handling event "{}", handler UID={}'.format(
                        event_name,
                        uid,
                    ),
                )

            if should_wait and future is not None:
                try:
                    future.result()
                except:
                    self.logger.exception(
                        "Error occurred while waiting for handler to finish "
                        '("{}") for event "{}"'.format(
                            event_handler.__name__,
                            event_name,
                        ),
                    )
                    pass

        return True



import subprocess


def check_ffmpeg_installed():
    """
    Check if ffmpeg is installed on the system.

    Returns:
        True if ffmpeg is installed, False otherwise.
    """
    try:
        subprocess.run(
            ["ffmpeg", "-version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except FileNotFoundError:
        return False
    return True


def convert_raw_video_to_output_file(
    raw_filepath: str,
    output_filepath: str,
    verbose=False,
):
    """
    Convert a raw video file to an output file using ffmpeg.

    Parameters:
        raw_filepath: The path to the raw video file.
        output_filepath: The path to the output file.
        verbose: Will output ffmpeg output to stdout
    """
    # Suppress ffmpeg output if not verbose
    stdout = subprocess.DEVNULL if not verbose else None

    subprocess.run(
        [
            "ffmpeg",
            "-i",
            raw_filepath,
            "-vf",
            "v360=dfisheye:e:ih_fov=193:iv_fov=193",
            "-y",
            output_filepath,
        ],
        stdout=stdout,
        stderr=subprocess.STDOUT,
    )
